#   Copyright 2012-2023 Darsh Ranjan, oblique
#
#   This file is part of termcolors.
#
#   termcolors is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   termcolors is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with termcolors.  If not, see
#   <http://www.gnu.org/licenses/>.

import os
import re
import select
import termios

from .colors import RGBAColor

#######################################################################
# Query-related error conditions


class TerminalQueryError(Exception):
    """
    Base class for the other exceptions.
    """

    def __init__(self, message):
        Exception.__init__(self, message)


class TerminalSetupError(TerminalQueryError):
    """
    We couldn't set up the terminal properly.
    """

    def __init__(self, fd):
        TerminalQueryError.__init__(
            self,
            ("Couldn't set up terminal on file " +
             ("descriptor %d" % fd)))


class NoResponseError(TerminalQueryError):
    """
    The terminal didn't respond, or we were too impatient.
    """

    def __init__(self, q):
        TerminalQueryError.__init__(
            self,
            "Timeout on query " + repr(q))


class TerminalUninitializedError(TerminalQueryError):
    """
    Someone tried to do something without setting up the terminal
    properly (by calling TerminalQueryContext.__enter__).
    """

    def __init__(self, fd):
        TerminalQueryError.__init__(
            self,
            (("Terminal on file descriptor %d " % fd) +
             "not set up"))


########################################################################

class TerminalQueryContext(object):
    """
    Context manager for terminal RGB queries.
    """

    # Operating system command
    osc = "\033]"

    # String terminator
    #  ("\033\\" is another option, but "\007" seems to be understood by
    #  more terminals.  Terminology, for example, doesn't seem to like
    #  "\033\\".)
    st = "\007"

    # Control sequence introducer
    csi = "\033["

    # ANSI SGR0
    reset = csi + 'm'

    def __init__(self, fd=0, screen_forward=False):
        """
        Arguments:
            fd: open file descriptor referring to the terminal we care
                about. The default (0) is almost always correct.
            screen_forward: whether to attempt to forward queries
                through a screen or tmux session if we are in one.
        """
        self.tc_save = None
        self.fd = fd

        self.num_errors = 0

        self.screen_forward = screen_forward

    def __enter__(self):
        """
        Set up the terminal for queries.
        """
        self.tc_save = termios.tcgetattr(self.fd)

        tc = termios.tcgetattr(self.fd)

        # Don't echo the terminal's responses
        tc[3] &= ~termios.ECHO

        # Noncanonical mode (i.e., disable buffering on the terminal
        # level)
        tc[3] &= ~termios.ICANON

        # Make input non-blocking
        tc[6][termios.VMIN] = 0
        tc[6][termios.VTIME] = 0

        termios.tcsetattr(self.fd, termios.TCSANOW, tc)

        # Check if it succeeded
        if termios.tcgetattr(self.fd) != tc:
            termios.tcsetattr(self.fd, termios.TCSANOW, self.tc_save)
            raise TerminalSetupError(self.fd)

        self.P = select.poll()
        self.P.register(self.fd, select.POLLIN)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Reset the terminal to its original state.
        """
        self.flush_input()

        if self.tc_save is not None:
            termios.tcsetattr(self.fd, termios.TCSANOW, self.tc_save)

        del self.P

    def get_num_colors(self, timeout=-1):
        """
        Attempt to determine the number of colors we are able to query
        from the terminal. A larger timeout is safer but will cause this
        function to take proportionally more time.

        Arguments:
            timeout: millisecond timeout, same interpretation as in
                self.guarded_query.

        Return: integer number of colors that are queryable.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        # We won't count failed queries in this function, since we're
        # guaranteed to fail a few.
        num_errors = self.num_errors

        if not self.get_indexed_color(0, timeout):
            return 0

        a = 0
        b = 1
        while self.get_indexed_color(b, timeout):
            a = b
            b += b

        while b - a > 1:
            c = (a + b) >> 1
            if self.get_indexed_color(c, timeout):
                a = c
            else:
                b = c

        self.num_errors = num_errors
        return b

    def get_all_indexed_colors(self, limit=-1, timeout=-1):
        """
        Query as many indexed RGB values as possible up to `limit`
        and return them all in a list.

        Arguments:
            limit: maximum number of colors to query, if nonnegative.
                Negative values disable the limit.
            timeout: millisecond timeout, same interpretation as in
                self.guarded_query.

        Return: list of RGBAColor instances.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        colors = []

        k = 0
        while limit < 0 or k < limit:
            c = self.get_indexed_color(k, timeout)
            if c:
                colors.append(c)
                k += 1
            else:
                if limit < 0:
                    self.num_errors -= 1
                break

        return colors

    # Wrappers for xterm & urxvt operating system controls.
    #
    # These codes are all common to xterm and urxvt. Their responses
    # aren't always in the same format (xterm generally being more
    # consistent), but the regular expression used to parse the
    # responses is general enough to work for both.
    #
    # Note: none of these functions is remotely thread-safe.

    def get_fg(self, timeout=-1):
        """
        Get the terminal's foreground (text) color.

        Arguments:
            timeout: millisecond timeout, same interpretation as in
                self.guarded_query.

        Return: RGBAColor instance.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        return self.rgb_query([10], timeout)

    def get_bg(self, timeout=-1):
        """
        Get the terminal's background color.

        Arguments:
            timeout: millisecond timeout, same interpretation as in
                self.guarded_query.

        Return: RGBAColor instance.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        return self.rgb_query([11], timeout)

    def get_indexed_color(self, a, timeout=-1):
        """
        Get color number `a`.

        Arguments:
            a: color index to query.
            timeout: millisecond timeout, same interpretation as in
                self.guarded_query.

        Return: RGBAColor instance.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        return self.rgb_query([4, a], timeout)

    def flush_input(self):
        """
        Discard any input that can be read at this moment.
        """
        while self.P.poll(0):
            os.read(self.fd, 4096)

    # Patterns matching unsigned decimal and hexadecimal integer
    # literals
    ndec = "[0-9]+"
    nhex = "[0-9a-fA-F]+"

    # The "guard" query and its response pattern.
    q_guard = csi + "6n"

    str_guard = "(.*)\033\\[{ndec};{ndec}R".format(**vars())
    re_guard = re.compile(str_guard)

    # This is what we expect the terminal's response to a query for a
    # color to look like.  If we didn't care about urxvt, we could get
    # away with a simpler implementation here, since xterm and vte seem
    # to give pretty consistent and systematic responses.
    str_rgb = ("\033\\]({ndec};)+rgba?:(({nhex})/)?" +
               "({nhex})/({nhex})/({nhex})").format(**vars())

    re_rgb = re.compile(str_rgb)

    def rgb_query(self, q, timeout=-1):
        r"""
        Query a color-valued terminal parameter.

        See
            http://invisible-island.net/xterm/ctlseqs/ctlseqs.html

        ("Operating System Controls") to see the various queries
        supported by xterm.  Urxvt supports some, but not all, of them,
        and has a number of its own (see man -s7 urxvt).

        self.__enter__ must be called prior to calling this function, or
        TerminalUninitializedError will be raised.

        Arguments:
            q: The query code as a sequence of nonnegative integers,
                i.e., [q0, q1, ...] if the escape sequence in
                pseudo-Python is

                    "\033]{q0};{q1};...;?\007"

            timeout: millisecond timeout, same interpretation as in
                self.guarded_query.

        Return: the color value as an RGBAColor instance. If the
            terminal provides an unparseable (or no) response, then None
            will be returned.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        query = (self.osc +
                 ';'.join([str(k) for k in q]) + ';?' +
                 self.st)

        try:
            response = self.guarded_query(query, timeout)
        except NoResponseError:
            return None

        m = self.re_rgb.match(response)

        if not m:
            self.num_errors += 1
            return None

        # (possibly overkill, since all terminals that reply seem to
        # give 4-digit RGB components, in which case `nd' is 4 and `u'
        # is 0xffff)
        nd = len(m.group(4))
        u = (1 << (nd << 2)) - 1

        # An "rgba"-type reply (for urxvt) is apparently actually
        #
        #    rgba:{alpha}/{alpha * red}/{alpha * green}/{alpha * blue}
        #
        # We opt to extract the actual RGB values by eliminating alpha.
        # (In other words, the alpha value is discarded completely in
        # the reported color value.)

        alpha = float(int(m.group(3), 16))/u if m.group(3) else 1.0

        if alpha > 0:
            (r, g, b) = (int(m.group(i), 16)/(alpha*u)
                         for i in [4, 5, 6])

            return RGBAColor(r, g, b, alpha)
        else:
            return RGBAColor(0.0, 0.0, 0.0, 0.0)

    # If a terminal sees an escape sequence it doesn't like, it will
    # simply ignore it. Also, it's hard to predict how long a terminal
    # will take to respond to a query it does like. However, some
    # escape sequences, like "\033[6n", will produce a predictable
    # response on *most* (but not all) terminals, and this fact can be
    # used to test for the absence of a response to a particular query
    # on such terminals.

    def guarded_query(self, q, timeout=-1, flush=True):
        """
        Send the terminal query string `q` and return the terminal's
        response.

        Arguments:
            q: the query string to send to the terminal.
            timeout: how many milliseconds to wait for a response, a
                negative number meaning "infinite".
            flush: whether to discard waiting input before sending the
                query.  It usually makes sense to do this, but note that
                the terminal may still send seemingly unsolicited data
                (possibly in response to an earlier query) after the
                input is flushed, so flushing the input does not
                provide any guarantees.

        Return: The terminal's response to the query as a string.

        Errors:
            NoResponseError: if the query times out.
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        if not hasattr(self, "P"):
            raise TerminalUninitializedError(self.fd)

        query = q + self.q_guard

        if self.screen_forward:
            if os.getenv("TMUX"):
                # We seem to be in a tmux session.  tmux can be used
                # like a proxy to send queries to the actual terminal.
                # The format of proxy query is:
                #
                #     \033Ptmux;{original query}\0\033\\
                #
                # Also, all the \033 of the original query must be
                # escaped with \033.
                query = ("\033Ptmux;" +
                         query.replace("\033", "\033\033") +
                         "\0\033\\")

            elif os.getenv("STY"):
                # We seem to be in a screen session.  Like tmux, screen
                # can also send queries through to the terminal.  Unlike
                # tmux, \033 does not need to be (cannot be???) escaped.
                #
                # XXX Is there a better way to "escape" string
                # terminators?
                query = ("\033P" +
                         query.replace("\033\\", "\033\033\\\033P\\") +
                         "\033\\")

        if flush:
            self.flush_input()

        os.write(self.fd, query.encode())

        response = ""

        while self.P.poll(timeout):
            response += os.read(self.fd, 4096).decode()
            m = self.re_guard.match(response)
            if m:
                return m.group(1)
        else:
            self.num_errors += 1
            raise NoResponseError(query)



#   Copyright 2012-2023 Darsh Ranjan
#
#   This file is part of termcolors.
#
#   termcolors is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   termcolors is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with termcolors.  If not, see
#   <http://www.gnu.org/licenses/>.

from .terminal_query import (TerminalQueryContext,
                             TerminalQueryError,
                             TerminalSetupError,
                             NoResponseError,
                             TerminalUninitializedError)

from .colors import RGBAColor



#   Copyright 2012-2023 Darsh Ranjan
#
#   This file is part of termcolors.
#
#   termcolors is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   termcolors is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with termcolors.  If not, see
#   <http://www.gnu.org/licenses/>.

from sys import (stdout, stderr)
from argparse import (ArgumentParser, ArgumentError)

from .color_display import ColorDisplay

########################################################################
# Command-line arguments

timeout_dft = -1

parser = ArgumentParser(
        description="Python script to show off terminal colors.",
        epilog="Run this script from the terminal whose colors " +
               "you want to showcase. " +
               "Only xterm-like terminals are supported " +
               "(see the README).")

mode_group = parser.add_mutually_exclusive_group()

p_choices = [16, 88, 256]

arg_p = mode_group.add_argument(
        '-p', '--pretty',
        action='store_true', default=False,
        help="show colors 0 through N-1 in a pretty format.  " +
             ("N must belong to %r.  " % p_choices) +
             "If N > 16, it should be the actual number of colors " +
             "supported by the terminal, or the output will almost " +
             "certainly not be pretty.")

mode_group.add_argument(
        '-f', '--flat',
        action='store_true', default=False,
        help="show a simple table with colors 0 through N-1.  ")

parser.add_argument(
        'n', nargs='?', metavar='N',
        type=int, default=16,
        help="number of colors to show.  " +
             "Unless you explicitly supply -p/--pretty or -f/--flat, " +
             "--pretty is used if possible and --flat is used " +
             "otherwise.  " +
             "N defaults to 16, showing the ANSI colors 0-15.  " +
             "If N is 0, the script will attempt to determine the " +
             "maximum number of colors automatically " +
             "(which may be slow).")

parser.add_argument(
        '--no-fgbg',
        action='store_false', dest='fgbg', default=True,
        help="suppress display of foreground/background colors.")

parser.add_argument(
        '--no-query',
        action='store_false', dest='do_query', default=True,
        help="don't try to query any RGB values from the terminal " +
             "and just use placeholders.")

parser.add_argument(
        '-t', '--timeout', metavar='T',
        type=int, default=timeout_dft,
        help="how long to wait for the terminal to "
             "respond to a query, in milliseconds  " +
             "[default: {0}].  ".format(timeout_dft) +
             "If your output has '?' characters " +
             "instead of RGB values " +
             "or junk printed after the script runs, " +
             "increasing this value may or may not " +
             "help, depending on the terminal.  " +
             "A negative T will behave like infinity.")

parser.add_argument(
        '-l', '--level', metavar='L',
        type=int, default=3,
        help="choose how much color to use in the output.  " +
             "(0 = no color; 3 = most color [default])")

parser.add_argument(
        '--screen-forward', '--tmux-forward',
        action='store_true', default=False,
        help="attempt to pass terminal queries through a tmux " +
             "session (ignored if this is not a tmux session)")

########################################################################


def main():
    args = parser.parse_args()

    assert not (args.pretty and args.flat)

    if args.pretty:
        if args.n not in p_choices:
            raise ArgumentError(
                    arg_p,
                    "N must belong to %r" % p_choices)

    with ColorDisplay(0, args.timeout, args.level, args.do_query,
                      args.screen_forward) as C:

        if args.n == 0:
            args.n = C.get_num_colors(args.timeout)

        if not (args.pretty or args.flat):
            if args.n in p_choices:
                args.pretty = True
            else:
                args.flat = True

        if args.level >= 1:
            stdout.write(C.reset)

        if args.fgbg:
            C.show_fgbg()

        if args.pretty:
            assert args.n in p_choices

            stdout.write('\n    ANSI colors:\n\n')
            C.show_ansi()

            if args.n > 16:
                stdout.write('\n    RGB cube:\n\n')
                C.show_color_cube(args.n)

                stdout.write('    Grayscale ramp:\n\n')
                C.show_grayscale_ramp(args.n)
        else:
            C.show_colors(args.n)

        if C.num_errors > 0:
            stderr.write("Warning: not all queries succeeded\n" +
                         "Warning:     (output contains " +
                         "placeholders and may be inaccurate)\n")


if __name__ == "__main__":
    main()



#   Copyright 2012-2023 Darsh Ranjan
#
#   This file is part of termcolors.
#
#   termcolors is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   termcolors is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with termcolors.  If not, see
#   <http://www.gnu.org/licenses/>.

from collections import defaultdict
from sys import stdout

from .terminal_query import TerminalQueryContext


class ColorDisplay(TerminalQueryContext):
    """
    Class for producing a colored display of terminal RGB values.  It's
    best to use this class as a context manager, which will properly set
    and reset the terminal's attributes.
    """

    def __init__(self, tty_fd,
                 timeout=100, color_level=3, do_query=True,
                 screen_forward=False):
        """
        Arguments:
            tty_fd: open file descriptor connected to a terminal.
            timeout: same interpretation as in rgb_query. A larger
                timeout will be used a small number of times to test the
                capabilities of the terminal.
            color_level: how much color should be in the output. Use 0
                to suppress all color and 3 or greater for maximum
                coloredness.
            do_query: whether to attempt to query RGB values from the
                terminal or just use placeholders everywhere.
            screen_forward: whether to attempt to forward queries
                through a screen or tmux session if we are in one.
        """
        TerminalQueryContext.__init__(self, tty_fd, screen_forward)

        self.timeout = timeout
        self.color_level = color_level
        self.do_query = do_query

        def none_factory():
            return None

        # colors for highlighting
        self.hi = defaultdict(none_factory)

        self.hi['['] = 10
        self.hi[']'] = 10
        self.hi['+'] = 9
        self.hi['/'] = 9

        for c in '0123456789ABCDEF':
            self.hi[c] = 12

        # String to use for color values that couldn't be determined
        self.rgb_placeholder = '??????'
        self.fmt = '{:02X}{:02X}{:02X}'
        self.scale = 0xff

    def __enter__(self):
        TerminalQueryContext.__enter__(self)

        # try getting the rgb value for color 0 to decide whether to
        # bother trying to query any more colors.
        self.do_query = (self.do_query and
                         self.get_indexed_color(0, self.timeout*2))

        if self.color_level >= 1:
            stdout.write(self.reset)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.color_level >= 1:
            stdout.write(self.reset)

        TerminalQueryContext.__exit__(self, exc_type, exc_value,
                                      traceback)

    def show_fgbg(self):
        """
        Show the foreground and background colors.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        if self.do_query:
            bg = self.format(self.get_bg(timeout=self.timeout))
            fg = self.format(self.get_fg(timeout=self.timeout))
        else:
            bg = self.rgb_placeholder
            fg = self.rgb_placeholder

        stdout.write("\n    Background: %s\n" % bg)
        stdout.write("    Foreground: %s\n\n" % fg)

    def show_ansi(self):
        """
        Show the 16 ANSI colors (colors 0-15).

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        color_order = [0, 1, 3, 2, 6, 4, 5, 7]

        names = ['   Black ', '    Red  ', '   Green ', '  Yellow ',
                 '   Blue  ', '  Magenta', '   Cyan  ', '   White ']

        stdout.write(self.fgcolor('15', 3))

        for k in range(8):
            a = color_order[k]
            stdout.write(names[a])

        stdout.write('\n')
        stdout.write(self.fgcolor(None, 3))

        c = None
        for k in range(8):
            a = color_order[k]
            c = self.hiprint('   [%X/%X] ' % (a, 8 + a), c)
        stdout.write('\n')

        self.show_color_table([0, 8], color_order)

    def show_color_cube(self, n):
        """
        Show the "RGB cube" (xterm colors 16-231 (256-color) or 16-79
        (88-color)).  The cube has sides of length 6 or 4 (for 256-color
        or 88-color, respectively).

        Arguments:
            n: 256 or 88.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        base = {256: 6, 88: 4}[n]

        c = None
        c = self.hiprint('[ + ]   ', c)
        for w in range(base):
            c = self.hiprint('[%X]      ' % w, c)
        stdout.write('\n\n' + self.fgcolor(None, 3))

        for u in range(base):
            for v in range(base):
                stdout.write(' '*v)

                x = (u*base + v)*base
                self.hiprint('  [%02X]  ' % (16 + x))
                stdout.write(self.fgcolor(None, 3))

                for w in range(base):
                    self.show_color(x + w + 16)
                stdout.write('\n')
            stdout.write('\n\n')

    def show_grayscale_ramp(self, end):
        """
        Show the "grayscale ramp" (xterm colors 232-255 (256-color) or
        80-87 (88-color)).

        Arguments:
            n: 256 or 88.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        start = {256: 232, 88: 80}[end]
        n = end - start

        vals = [self.get_color(a) for a in range(start, end)]

        c = None

        c = self.hiprint('[ ', c)
        for v in range(n):
            c = self.hiprint('%02X ' % (start + v), c)
        c = self.hiprint(']\n', c)

        stdout.write('\n ' + self.fgcolor(None, 3))

        for v in range(n):
            stdout.write(' ' + self.block(start + v, 2))
        stdout.write('\n ')

        for u in range(3):
            for v in range(n):
                stdout.write(' ')
                stdout.write(self.fgcolor(start + v, 2))
                stdout.write(vals[v][2*u: 2*(u + 1)])
                stdout.write(self.fgcolor(None, 2))
            stdout.write('\n ')
        stdout.write('\n')

    def show_colors(self, n):
        """
        Make a table showing colors 0 through n-1.

        Arguments:
            n: the number (int) of colors to show.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        self.show_color_table(range(0, n, 8), range(8), n, True)

    def show_color_table(self, rows, cols, stop=-1, label=False):
        """
        Make a color table with all possible color indices of the form
        rows[k] + cols[j] that are less than `stop` (if `stop` is not
        negative). If label is True, then print row and column labels.

        Arguments:
            rows, cols: iterable of int..
            stop: if nonnegative, the upper bound (non-inclusive) on the
                color indices to display.
            label: whether to show row and column labels.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        if label:
            self.hiprint('[ + ]')
            stdout.write(self.fgcolor(None, 3))

            for a in cols:
                stdout.write('   ' + self.octal(a) + '  ')
            stdout.write('\n' + self.fgcolor(None, 1))

        if label:
            stdout.write('     ')

        stdout.write('\n')

        for b in rows:
            if label:
                stdout.write(self.octal(b) + ' ' +
                             self.fgcolor(None, 1))

            for a in cols:
                c = a + b
                if stop < 0 or c < stop:
                    self.show_color(b + a)
                else:
                    stdout.write('         ')
            stdout.write('\n')
        stdout.write('\n')

    def show_color(self, a):
        """
        Make a pretty display of color number `a`, showing a block of
        that color followed by the 6-character hexadecimal code for the
        color.

        Arguments:
            a: color index to show.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        stdout.write(' ' + self.block(a) + ' ')
        stdout.write(self.fgcolor(a, 2) + (self.get_color(a)))
        stdout.write(self.fgcolor(None, 2))

    def hiprint(self, s, last_color=-1):
        """
        Print s to stdout, highlighting digits, brackets, etc. if the
        color level allows it.

        Arguments:
            s: the string to print.
            last_color: the current terminal foreground color.  This
                should be `None` if no color is set, or the current
                color index, or something else (like a negative integer)
                if the color isn't known. (The last option is always
                safe and will force this function to do the right
                thing.)

        Return: the current foreground color, which can be passed as
            last_color to the next call if the color isn't changed in
            between.
        """
        for c in s:
            if c == ' ':
                color = last_color
            else:
                color = self.hi[c]

            if color != last_color:
                stdout.write(self.fgcolor(color, 3))

            stdout.write(c)
            last_color = color

        return last_color

    def octal(self, x):
        """
        Return a base-8 string for the integer x, highlighted if the
        color level allows it.

        Arguments:
            x: integer to convert.

        Return: string representation, possibly with ANSI color codes.
        """
        return (self.fgcolor(self.hi['+'], 3) + '0'
                + self.fgcolor(self.hi['0'], 3) + ('%03o' % x))

    def block(self, c, n=1):
        """
        Return a string that prints as a block of color `c` and size `n`.

        Arguments:
            c: color index of block.
            n: length of block.

        Return: string representation, possibly with ANSI color codes.
        """
        return self.bgcolor(c, 1) + ' '*n + self.bgcolor(None, 1)

    # Changing the foreground and background colors.
    #
    # While the 38;5 and 48;5 SGR codes are less portable than the usual
    # 30-37 and 40-47, these codes seem to be fairly widely implemented
    # (on X-windows terminals, screen, and tmux) and support the whole
    # color range, as opposed to just colors 0-8.  They also make it
    # very easy to set the background to a given color without needing
    # to mess around with bold or reverse video (which are hardly
    # portable themselves).  This is useful even for the 16 ANSI colors.

    def fgcolor(self, a=None, level=-1):
        """
        Return a string designed to set the foreground color to `a` when
        printed to the terminal. None means default.

        Arguments:
            a: color index to set to, or None to reset to default.
            level: minimum colorfulness level for which colors should be
                affected.

        Return: ANSI control sequence as string.
        """
        if self.color_level >= level:
            if a is None:
                return self.csi + '39m'
            else:
                return self.csi + '38;5;' + str(a) + 'm'
        else:
            return ''

    def bgcolor(self, a=None, level=-1):
        """
        Return a string designed to set the background color to `a` when
        printed to the terminal. None means default.

        Arguments:
            a: color index to set to, or None to reset to default.
            level: minimum colorfulness level for which colors should be
                affected.

        Return: ANSI control sequence as string.
        """
        if self.color_level >= level:
            if a is None:
                return self.csi + '49m'
            else:
                return self.csi + '48;5;' + str(a) + 'm'
        else:
            return ''

    def get_color(self, a):
        """
        Return a formatted string representing the given color index,
        if possible.

        Arguments:
            a: color index to convert.

        Return: the color's hex code, or a placeholder.

        Errors:
            TerminalUninitializedError: if this instance's context has
                not been entered.
        """
        if self.do_query:
            c = self.get_indexed_color(a, timeout=self.timeout)
            return self.format(c)
        else:
            return self.rgb_placeholder

    def format(self, c):
        """
        Return a formatted string representing RGBAColor instance c.

        Arguments:
            c: RGBAColor instance to convert, or None.

        Return: hex code of c, or a placeholder.
        """
        if c:
            return "%02X%02X%02X" % (int(c.r * 0xffff) // 256,
                                     int(c.g * 0xffff) // 256,
                                     int(c.b * 0xffff) // 256)
        else:
            return self.rgb_placeholder
